# ---------------------------------------------------------------------------- #
# paramorse.core.payload
# ---------------------------------------------------------------------------- #

from __future__ import annotations

import copy
import logging

from typing import Union
from typing import TypedDict
from typing import Literal
from typing import TypeAlias
from typing import cast
from typing import get_args

from paramorse.utils.render import pm_pformat # noqa: F401 

import paramorse.utils.helpers as pm_helpers
from paramorse.core.config import ParaMorseConfig

logger = logging.getLogger(__name__)

payload_input_type_alias: TypeAlias = Union[
    str,
    list[str],
    list[list[str]],
]

# ---------------------------------------------------------------------------- #

PayFormat_Literal = Literal["string", "string_list", "nest_list"]

PayKind_Literal = Literal["dd", "sym", "para"]

class PM_Payload(TypedDict):
    sym: str
    sym_list: list[str]

    dd: str
    dd_list: list[str]
    dd_flat: list[str]
    dd_nest: list[list[str]]

    para: str
    para_list: list[str]
    para_flat: list[str]
    para_flat_rs: list[str] # right space
    para_nest: list[list[str]]

    pay_format: PayFormat_Literal | None
    pay_kind: PayKind_Literal | None


# make literals into tuples for validity checks 
PAY_FORMATS = cast(tuple[PayFormat_Literal, ...], get_args(PayFormat_Literal))
PAY_KINDS = cast(tuple[PayKind_Literal, ...], get_args(PayKind_Literal))

_PM_PAYLOADS: dict[str, PM_Payload] = {
    "empty": PM_Payload(
        sym="",
        sym_list=[],
        dd="",
        dd_list=[],
        dd_flat=[],
        dd_nest=[],
        para="",
        para_list=[],
        para_flat=[],
        para_flat_rs=[],
        para_nest=[],
        pay_format=None,
        pay_kind=None,
    )
}


# ---------------------------------------------------------------------------- #
# MORSE MAPS

# between kinds:
# dd: dotdash
# sym: symbol
# para: paralanguage

# across formats:
# string: str
# list: list of str
# nest: list of list of str


def morse_sym_to_dd(
        symbols: str,
        variant_cfg_input: dict | None = None,
    ) -> str:

    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()
    morse_alphabet = variant_cfg["symbol_to_dd"]

    symbols = symbols.upper()
    dds: list[str] = []
    for sym in symbols:
        if sym == " ":
            dds.append("/")
        elif sym in morse_alphabet:
            dds.append(morse_alphabet[sym])
        else:
            logger.warning("No Morse for %r", sym)

    morse_dd = " ".join(dds)

    return morse_dd


def morse_dd_to_sym(
        morse_dd: str,
        variant_cfg_input: dict | None = None,
    ) -> str:
    
    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()
    rev_morse_alphabet = variant_cfg["dd_to_symbol"]

    dd_split = morse_dd.split()
    symbols: list[str] = []
    for dd in dd_split:
        if dd == "/":
            symbols.append(" ")
        elif dd in rev_morse_alphabet:
            symbols.append(rev_morse_alphabet[dd])
        else:
            logger.warning("Unknown Morse %r", dd)

    morse_sym = "".join(symbols)

    return morse_sym


def morse_dd_to_para_nest(
        morse_dd: str,
        variant_cfg_input: dict | None = None,
    ) -> list:

    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    # assumes dotdashes of letters are space separated
    dd_split = morse_dd.split()
    para_nest = []
    for i_dd, dd in enumerate(dd_split):
        # find seg in alphabet
        if dd in variant_cfg["dd_to_symbol"]:
            morse_sym = variant_cfg["dd_to_symbol"][dd]
            morse_para = variant_cfg["symbol_to_para"][morse_sym]
            para_nest.append(morse_para)

        # skip paralanguage separators on last morse symbol
        if (i_dd + 1) == len(dd_split):
            continue

        # new word separator, if any
        if dd == "/":
            if variant_cfg["word_sep_overwrite"]:
                para_nest.pop()  # drop last end of symbol letter_separator
            para_nest.append([variant_cfg["word_separator"]])

        # new letter separator, if any
        else:
            para_nest.append([variant_cfg["letter_separator"]])

    return para_nest


def morse_para_expand(
        morse_para: str,
        variant_cfg_input: dict | None = None,
    ) -> dict[str, list]:
    
    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    # assumes morse_para tokens are space separated
    para_split = morse_para.split()
    para_top_list = []
    para_cur_sublist = []
    for para in para_split:
        if para == variant_cfg["dot_token"]:
            para_cur_sublist.append(para)
        elif para == variant_cfg["dash_token"]:
            para_cur_sublist.append(para)
        elif (
            para == variant_cfg["letter_separator"]
            or para == variant_cfg["word_separator"]
        ):
            if para_cur_sublist:
                para_top_list.append(para_cur_sublist)
            para_top_list.append([para])
            para_cur_sublist = []
        else:
            logger.warning("No Morse for %r", para)

    # flush last segment
    if para_cur_sublist:
        para_top_list.append(para_cur_sublist)

    para_expansion = {
        "para_list": para_split,
        "para_nest": para_top_list,
    }

    return para_expansion


def morse_para_to_dd(
        morse_para: str,
        variant_cfg_input: dict | None = None,
    ) -> str:

    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    # assumes morse_para tokens are space separated
    para_split = morse_para.split()
    dds: list[str] = []
    for para in para_split:
        if para == variant_cfg["dot_token"]:
            dds.append(".")
        elif para == variant_cfg["dash_token"]:
            dds.append("-")
        elif para == variant_cfg["letter_separator"]:
            dds.append(" ")
        elif para == variant_cfg["word_separator"]:
            if variant_cfg["word_sep_overwrite"]:
                dds.append(" ")
                dds.append("/")
            else:
                dds.append("/")
            dds.append(" ")
        else:
            logger.warning("No Morse for %r", para)

    morse_dd = "".join(dds)

    return morse_dd


# ---------------------------------------------------------------------------- #
# DETECT PAYLOAD
# payload kind 
# payload format

def _detect_payload_format(
        payload_input: payload_input_type_alias,
    ) -> PayFormat_Literal | None:

    pay_format = None

    if pm_helpers.is_string(payload_input):
        pay_format = "string"

    elif pm_helpers.is_list_of_strings(payload_input):
        pay_format = "string_list"

    elif pm_helpers.is_list_of_list_of_strings(payload_input):
        pay_format = "nest_list"

    return pay_format


def _detect_payload_kind(
        payload_input: payload_input_type_alias,
        pay_format: str | None = None,
        variant_cfg_input: dict | None = None,
    ) -> PayKind_Literal | None:
    
    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    dd_para_map: dict = variant_cfg["dd_para_map"]
    tok_map_dd_set = set(dd_para_map.keys())
    tok_map_para_set = set(dd_para_map.values())

    pay_kind = None

    if payload_input is None or (
        not isinstance(payload_input, str) and not isinstance(payload_input, list)
    ):
        return pay_kind

    # return on empty input
    payload_input_len = len(payload_input)
    if not payload_input_len > 0:
        return pay_kind

    if pay_format not in PAY_FORMATS:
        pay_format = _detect_payload_format(payload_input)

    if pay_format == "string":
        if not pm_helpers.is_string(payload_input):
            raise TypeError(f"""Expected str payload for pay_format '{pay_format}'""")
        
        payload_set = set(payload_input)
        payload_split_set = set(payload_input.split())

        if payload_set.issubset(tok_map_dd_set):
            pay_kind = "dd"
        elif payload_split_set.issubset(tok_map_para_set):
            pay_kind = "para"
        else:
            pay_kind = "sym"

    elif pay_format == "nest_list":
        if not pm_helpers.is_list_of_list_of_strings(payload_input):
            raise TypeError(f"""Expected list[list[str]] payload for pay_format '{pay_format}'""")
        
        payload_input_set = set()
        for pld_list in payload_input:
            payload_input_set = payload_input_set.union(set(pld_list))

        if payload_input_set.issubset(tok_map_dd_set):
            pay_kind = "dd"
        elif payload_input_set.issubset(tok_map_para_set):
            pay_kind = "para"
        else:
            pay_kind = "sym"

    elif pay_format == "string_list":
        if not pm_helpers.is_list_of_strings(payload_input):
            raise TypeError(f"""Expected list[str] payload for pay_format '{pay_format}'""")
        
        payload_input_set = set()
        payload_input_split_set = set()
        for pld_str in payload_input:
            payload_input_set = payload_input_set.union(set(pld_str))
            payload_input_split_set = payload_input_split_set.union(
                set(pld_str.split())
            )
        if payload_input_set.issubset(tok_map_dd_set):
            pay_kind = "dd"
        elif payload_input_split_set.issubset(tok_map_para_set):
            pay_kind = "para"
        else:
            pay_kind = "sym"

    # logger.debug(f'pay_kind: {pay_kind}')

    return pay_kind


def detect_payload(
        payload_input: payload_input_type_alias,
        variant_cfg_input: dict | None = None,
    ) -> dict:

    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    pay_format = _detect_payload_format(payload_input)
    pay_kind = _detect_payload_kind(
        payload_input=payload_input,
        pay_format=pay_format,
        variant_cfg_input=variant_cfg,
    )

    # logger.debug(f"pay_format: {pay_format}")
    # logger.debug(f"pay_kind: {pay_kind}")

    detect_result = {
        "pay_format": pay_format,
        "pay_kind": pay_kind,
    }

    return detect_result


# ---------------------------------------------------------------------------- #
# CREATE PAYLOAD


def _start_payload(
        Pay_Dict: PM_Payload,
        payload_input: payload_input_type_alias,
        pay_format: str,
        pay_kind: str,
        variant_cfg_input: dict | None = None,
    ) -> PM_Payload:
    
    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    if pay_format == "string":
        if not pm_helpers.is_string(payload_input): return Pay_Dict

        payload_input = payload_input.strip()

        if pay_kind == "dd":
            Pay_Dict["dd"] = payload_input
            Pay_Dict["dd_list"] = Pay_Dict["dd"].split(" ")
            Pay_Dict["dd_nest"] = [list(mark) for mark in Pay_Dict["dd_list"]]
        elif pay_kind == "sym":
            Pay_Dict["sym"] = payload_input
            Pay_Dict["sym_list"] = list(Pay_Dict["sym"].split())
        elif pay_kind == "para":
            Pay_Dict["para"] = payload_input
            para_expansion = morse_para_expand(Pay_Dict["para"], variant_cfg)
            Pay_Dict["para_list"] = para_expansion["para_list"]
            Pay_Dict["para_nest"] = para_expansion["para_nest"]

    elif pay_format == "string_list":
        if not pm_helpers.is_list_of_strings(payload_input): return Pay_Dict

        if pay_kind == "dd":
            Pay_Dict["dd_list"] = payload_input
            Pay_Dict["dd"] = " ".join(Pay_Dict["dd_list"])
            Pay_Dict["dd_nest"] = [list(mark) for mark in Pay_Dict["dd_list"]]
        elif pay_kind == "sym":
            Pay_Dict["sym_list"] = payload_input
            Pay_Dict["sym"] = " ".join(Pay_Dict["sym_list"])
        elif pay_kind == "para":
            Pay_Dict["para_list"] = payload_input
            Pay_Dict["para_nest"] = [mark.split() for mark in payload_input]
            Pay_Dict["para"] = " ".join(Pay_Dict["para_list"])

    elif pay_format == "nest_list":
        if not pm_helpers.is_list_of_list_of_strings(payload_input): return Pay_Dict

        if pay_kind == "dd":
            Pay_Dict["dd_nest"] = payload_input
            Pay_Dict["dd_list"] = ["".join(mark) for mark in Pay_Dict["dd_nest"]]
            Pay_Dict["dd"] = " ".join(Pay_Dict["dd_list"])
        elif pay_kind == "para":
            Pay_Dict["para_nest"] = payload_input
            Pay_Dict["para_list"] = [" ".join(mark) for mark in Pay_Dict["para_nest"]]
            Pay_Dict["para"] = " ".join(Pay_Dict["para_list"])

    return Pay_Dict


def _fill_out_payload(
        Pay_Dict: PM_Payload, pay_kind: str, variant_cfg_input
    ) -> PM_Payload:

    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    if pay_kind == "dd":
        Pay_Dict["sym"] = morse_dd_to_sym(Pay_Dict["dd"], variant_cfg)
        Pay_Dict["sym_list"] = list(Pay_Dict["sym"].split())
        Pay_Dict["para_nest"] = morse_dd_to_para_nest(Pay_Dict["dd"], variant_cfg)
        Pay_Dict["para_list"] = [para for mark in Pay_Dict["para_nest"] for para in mark]
        Pay_Dict["para"] = " ".join(Pay_Dict["para_list"])

    elif pay_kind == "sym":
        Pay_Dict["dd"] = morse_sym_to_dd(Pay_Dict["sym"], variant_cfg)
        Pay_Dict["dd_list"] = Pay_Dict["dd"].split(" ")
        Pay_Dict["dd_nest"] = [list(mark) for mark in Pay_Dict["dd_list"]]
        Pay_Dict["para_nest"] = morse_dd_to_para_nest(Pay_Dict["dd"], variant_cfg)
        Pay_Dict["para_list"] = [" ".join(mark) for mark in Pay_Dict["para_nest"]]
        Pay_Dict["para"] = " ".join(Pay_Dict["para_list"])

    elif pay_kind == "para":
        Pay_Dict["dd"] = morse_para_to_dd(Pay_Dict["para"], variant_cfg)
        Pay_Dict["dd_list"] = Pay_Dict["dd"].split(" ")
        Pay_Dict["dd_nest"] = [list(mark) for mark in Pay_Dict["dd_list"]]
        Pay_Dict["sym"] = morse_dd_to_sym(Pay_Dict["dd"], variant_cfg)
        Pay_Dict["sym_list"] = list(Pay_Dict["sym"].split())

    # flat dd lists
    Pay_Dict["dd_flat"] = []
    for mark in Pay_Dict["dd"].split():
        Pay_Dict["dd_flat"].extend(mark)
        Pay_Dict["dd_flat"].append(" ")

    # flat para lists
    Pay_Dict["para_flat"] = Pay_Dict["para"].split()
    Pay_Dict["para_flat_rs"] = [para + " " for para in Pay_Dict["para_flat"]]

    return Pay_Dict


def build_payload(
        payload_input: payload_input_type_alias,
        pay_config: dict | None = None,
        variant_cfg_input: dict | None = None,
    ) -> PM_Payload:

    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    Pay_Dict: PM_Payload = copy.deepcopy(_PM_PAYLOADS["empty"])

    if not (
        pm_helpers.is_string(payload_input)
        or pm_helpers.is_list_of_strings(payload_input)
        or pm_helpers.is_list_of_list_of_strings(payload_input)
    ):
        return Pay_Dict

    # SET pay_format & pay_kind

    if pay_config is not None:
        pay_format = pay_config.get("pay_format", None)
        pay_kind = pay_config.get("pay_kind", None)

    pay_format = _detect_payload_format(payload_input)

    pay_kind = _detect_payload_kind(
        payload_input=payload_input,
        pay_format=pay_format,
        variant_cfg_input=variant_cfg,
    )

    Pay_Dict["pay_format"] = pay_format
    Pay_Dict["pay_kind"] = pay_kind

    if pay_format not in PAY_FORMATS or pay_kind not in PAY_KINDS:
        return Pay_Dict

    # BUILD Pay_Dict

    _start_payload(
        Pay_Dict=Pay_Dict,
        payload_input=payload_input,
        pay_format=pay_format,
        pay_kind=pay_kind,
        variant_cfg_input=variant_cfg,
    )

    _fill_out_payload(
        Pay_Dict=Pay_Dict,
        pay_kind=pay_kind,
        variant_cfg_input=variant_cfg,
    )

    # logger.debug(f"Pay_Dict:\n{pm_pformat(Pay_Dict)}")

    return Pay_Dict



# ---------------------------------------------------------------------------- #
# paramorse.utils.paths
# ---------------------------------------------------------------------------- #

from __future__ import annotations

from pprint import PrettyPrinter
from pathlib import Path
from typing import Any
import os
import logging

_pp = PrettyPrinter(
    indent=2,
    width=80,
    compact=True,
    # compact=False,
    # sort_dicts=True,
    sort_dicts=False,
)

logger = logging.getLogger(__name__)

_ENV_PM_DATADIR = "PARAMORSE_DATA_DIR"

_ROOT_DATADIR_NAME = "pm_data"

_PM_DATADIR_PATH = Path(
    os.getenv(
        _ENV_PM_DATADIR,
        Path.cwd() / _ROOT_DATADIR_NAME
        )
    ).expanduser().resolve()

if(not _PM_DATADIR_PATH.is_dir):
    _PM_DATADIR_PATH = Path.cwd() / _ROOT_DATADIR_NAME

_DATADIR_NAMES = {
    # <data>: _ROOT_DATADIR_NAME/<dirname>
    "root": "",
    "raw": "raw",
    "cover": "cover",
    "payload": "payload",
    "encoded_pkg": "package_enc",
    "decoded_pkg": "package_dec",
    "tts": "tts",
    "transcribe": "transcribe",
    "test": "test",
    "misc": "misc",
}

_DATADIR_PATH_OBJS = dict()
for data, dirname in _DATADIR_NAMES.items():
    _DATADIR_PATH_OBJS[data] = (Path(_PM_DATADIR_PATH) / dirname).expanduser().resolve()

_DATADIR_PATH_STRINGS = dict()
for data, path_obj in _DATADIR_PATH_OBJS.items():
    _DATADIR_PATH_STRINGS[data] = str(path_obj)

def get_data_dirs() -> dict:
    data_dirs_dict = dict()
    data_dirs_dict['names'] = _DATADIR_NAMES
    data_dirs_dict['paths_obj'] = _DATADIR_PATH_OBJS
    data_dirs_dict['paths_string'] = _DATADIR_PATH_STRINGS
    return data_dirs_dict 

def _ensure_dir(dir_path: Path) -> Path:
    dir_path.mkdir(parents=True, exist_ok=True)
    return dir_path

def _ensure_file(file_path: Path) -> Path:
    file_path.parent.mkdir(parents=True, exist_ok=True)
    file_path.touch()
    return file_path

def data_path_resolver(
        data_file_path: str | Path | None=None, # e.g. "mydir/myfile.txt"
        data_dir_name: str = _DATADIR_NAMES['misc'], # e.g. "transcribe"   
        data_item_name: str | None=None, # e.g. "peter_test"
        data_item_subname: str | None=None, # e.g. "dec", "enc", "whisper_transcribe" 
        data_item_id: str | None=None, # e.g. timestamp_string, uuid
        data_file_suffix: str | None=None, # e.g. ".mp3", "_tts.mp3"
    ) -> dict[str, Any]:

    resolve_path_dict = {}

    resolve_source = ""

    if data_file_path is not None:
        resolve_source = 'data_file_path'

        if(isinstance(data_file_path, str)):
            file_path_obj = Path(data_file_path).expanduser().resolve()
        elif(isinstance(data_file_path, Path)):
            file_path_obj = data_file_path.expanduser().resolve()
        else:
            logger.error(f'data_file_path must be str or Path, not {type(data_file_path)}:\n{data_file_path}')
            return resolve_path_dict
        if(file_path_obj.is_dir()):
            logger.error(f'file_path_obj is dir, should be file instead:\n{file_path_obj}')
            return resolve_path_dict
        dir_path_obj = file_path_obj.parent
        _ensure_dir(dir_path_obj)

    else:
        resolve_source = 'data_dir_name'

        if (data_item_name is None or data_item_name == ""): data_item_name = "noname"

        if (data_item_subname is None): data_item_subname = ""
        else: data_item_subname = "_" + data_item_subname

        if (data_item_id is None): data_item_id = ""
        else: data_item_id = "_" + data_item_id

        if data_file_suffix is None: data_file_suffix = ""

        dir_path_obj = _PM_DATADIR_PATH / data_dir_name / data_item_name 
        dir_path_obj = Path(dir_path_obj).expanduser().resolve()
        _ensure_dir(dir_path_obj)

        file_stem = data_item_name + data_item_subname + data_item_id
        file_name = file_stem + data_file_suffix
        file_path_obj = _PM_DATADIR_PATH / data_dir_name / data_item_name / file_name
        file_path_obj = Path(file_path_obj).expanduser().resolve()


    dir_path_string = str(dir_path_obj)
    file_path_string = str(file_path_obj)

    resolve_path_dict = {

        'resolve_source': resolve_source,
        'data_file_path': data_file_path,
        'data_dir_name': data_dir_name,

        'dir_path_obj': dir_path_obj,

        'file_path_obj' : file_path_obj,
        'file_path_obj_name' : file_path_obj.name,
        'file_path_obj_stem' : file_path_obj.stem,
        'file_path_obj_suffix' : file_path_obj.suffix,

        'dir_path_string' : dir_path_string,
        'file_path_string': file_path_string,
    }

    # logger.debug(f'resolve_path_dict:\n{_pp.pformat(resolve_path_dict, sort_dicts=False)}')

    return resolve_path_dict




# ---------------------------------------------------------------------------- #
# paramorse.utils.helpers
# ---------------------------------------------------------------------------- #

from __future__ import annotations

import os
import logging

from datetime import datetime
from uuid import UUID, uuid4
from typing import Any 
from typing import TypeGuard

from spacy.tokens.token import Token as spacyToken
from spacy.tokens.doc import Doc as spacyDoc

from paramorse.utils.paths import get_data_dirs

logger = logging.getLogger(__name__)


_DATADIRS: dict[str, dict]=get_data_dirs()


# FILE IO

def read_text_file(text_input_file: str) -> str:
    
    logger.debug(f"text_input_file:\n{text_input_file}")

    if os.path.isfile(text_input_file):
        text_input_filepath = text_input_file
    elif os.path.isfile(f"{_DATADIRS['paths_string']['raw']}/{text_input_file}"):
        text_input_filepath = f"{_DATADIRS['paths_string']['raw']}/{text_input_file}"
    elif os.path.isdir(text_input_file):
        logger.error("text_input_file provided is directory")
        return ""
    else:
        logger.error(f"text_input_file not found:\n{text_input_file}")
        return ""
    
    with open(text_input_filepath, 'r', encoding = 'utf8') as file:
        text = file.read()

    return text

def _split_file_ext(file: str) -> tuple[str, str]:
    file_name, file_ext = os.path.splitext(file)
    return file_name, file_ext




# GENERATE IDENTIFIERS

def get_timestamp_string(timestamp_format: str | None=None) -> str:
   
    time_now = datetime.now()

    if(timestamp_format is not None):
        timestamp_string = time_now.strftime(f"{timestamp_format}")
    else:
        # timestamp_string = time_now.strftime("%Y%m%d_%H_%M_%S")
        # timestamp_string = time_now.strftime("%Y%m%d_%H_%M_%S.%f")
        frac_seconds = time_now.strftime("%f")[0:3]
        timestamp_string = time_now.strftime(f"%Y%m%d_%H_%M_%S_{frac_seconds}")

    return timestamp_string


def _get_uuid(uuid_format: str = "string") -> str | UUID:
   
    uuid_object = uuid4()
    uuid_string = str(uuid_object)

    if(uuid_format=="string"):
        uuid_result = uuid_string
    else:
        uuid_result = uuid_object

    return uuid_result




# TYPE CHECKING

# STR

def is_string(input: Any) -> TypeGuard[str]:
    return isinstance(input, str)

def is_empty_string(input: Any) -> TypeGuard[str]:
    return isinstance(input, str) and input == ""


# SPACY

def is_spacyToken(input: Any) -> TypeGuard[spacyToken]:
    return isinstance(input, spacyToken)

def is_spacyDoc(input: Any) -> TypeGuard[spacyDoc]:
    return isinstance(input, spacyDoc)


# DICT

def is_dict(input: Any) -> TypeGuard[dict]:
    return isinstance(input, dict)

def is_empty_dict(input: Any) -> TypeGuard[dict]:
    return isinstance(input, dict) and len(input) == 0

def is_nonempty_dict(input: Any) -> TypeGuard[dict]:
    return isinstance(input, dict) and len(input) > 0


# LIST

def is_list(input: Any) -> TypeGuard[list]:
    return isinstance(input, list)

def is_empty_list(input: Any) -> TypeGuard[list[Any]]:
    return isinstance(input, list) and len(input) == 0

def is_nonempty_list(input: Any) -> TypeGuard[list[Any]]:
    return isinstance(input, list) and len(input) > 0

def is_list_of_strings(input: Any) -> TypeGuard[list[str]]:
    return isinstance(input, list) and all(isinstance(sub, str) for sub in input)


def is_nonempty_list_of_strings(input: Any) -> TypeGuard[list[str]]:
    return (
        isinstance(input, list)
        and len(input) > 0
        and all(isinstance(sub, str) for sub in input)
        )

def is_list_of_lists(input: Any) -> TypeGuard[list[list]]:
    return isinstance(input, list) and all(isinstance(sub, list) for sub in input)

def is_list_of_list_of_strings(input: Any) -> TypeGuard[list[list[str]]]:
    return (
        isinstance(input, list)
        and all(isinstance(sub, list) for sub in input)
        and all(all(isinstance(sub_ele, str) for sub_ele in sub) for sub in input)
    )

def is_list_of_nonempty_list_of_strings(input: Any) -> TypeGuard[list[list[str]]]:
    return (
        # outer list reqs:
        isinstance(input, list)
        and len(input) > 0                                 
        and all(
            # sublist reqs:
            isinstance(sub, list)
            and len(sub) > 0
            # sublist string reqs:
            and all(isinstance(sub_ele, str) for sub_ele in sub)
            for sub in input
        )
    )

def is_list_of_spacyTokens(input: Any) -> TypeGuard[list[spacyToken]]:
    return isinstance(input, list) and all(isinstance(sub, spacyToken) for sub in input)

def is_non_empty_list_of_spacyTokens(input: Any) -> TypeGuard[list[spacyToken]]:
    return (
        isinstance(input, list)
        and len(input) > 0
        and all(isinstance(sub, spacyToken) for sub in input)
    )




# ---------------------------------------------------------------------------- #
# paramorse.core.cover
# ---------------------------------------------------------------------------- #

from __future__ import annotations

import copy
import logging

from typing import Union
from typing import TypedDict
from typing import TypeAlias

from paramorse.utils.render import pm_pformat # noqa: F401 

from spacy.tokens.token import Token as spacyToken
from spacy.tokens.doc import Doc as spacyDoc

import paramorse.utils.helpers as pm_helpers
import paramorse.core.linguistics as pm_ling


from paramorse.core.config import ParaMorseConfig

logger = logging.getLogger(__name__)



cover_input_type_alias: TypeAlias = Union[
        str,
        list[str],
        list[spacyToken],
        spacyDoc,
    ]

# ---------------------------------------------------------------------------- #

class PM_Cover(TypedDict):
    cover_text: str
    cover_parse_sdoc: spacyDoc | None
    cover_parse_stoks: list[spacyToken]
    cover_parse_list: list[str]
    cover_parse_list_ws: list[str]
    cover_parse_cap: list[int]
    cover_list: list[str]
    cover_tokens: list[str]
    cover_tokens_cap: list[int]
    cover_tokens_stok_subsets: list[list[int]] 


_PM_COVERS: dict[str, PM_Cover] = {
    'empty': PM_Cover(
        cover_text = "",
        cover_parse_sdoc = None,
        cover_parse_stoks = [],
        cover_parse_list = [],
        cover_parse_list_ws = [],
        cover_parse_cap = [],
        cover_list = [],
        cover_tokens = [],
        cover_tokens_cap = [],
        cover_tokens_stok_subsets = [],
    )
}

# ---------------------------------------------------------------------------- #

def tokenize_cover(
        cover_text: str | None=None,
        variant_cfg_input: dict | None=None,
    ) -> PM_Cover:

    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    Cvr_Dict: PM_Cover = copy.deepcopy(_PM_COVERS['empty'])

    if(not isinstance(cover_text, str) or cover_text==""): return Cvr_Dict
    
    Cvr_Dict['cover_text'] = cover_text

    snlp = pm_ling.load_spacy_model(variant_cfg["spacy_model"])
    
    Cvr_Dict['cover_parse_sdoc'] = snlp(cover_text)
    Cvr_Dict['cover_parse_stoks'] = list(Cvr_Dict['cover_parse_sdoc'])
    Cvr_Dict['cover_parse_list'] = [stok.text for stok in Cvr_Dict['cover_parse_stoks']]
    Cvr_Dict['cover_parse_list_ws'] = [stok.text_with_ws for stok in Cvr_Dict['cover_parse_stoks']]
    
    # start_on_non_encoding_ele = False
    for i_stok, stok in enumerate(Cvr_Dict['cover_parse_stoks']):

        if(pm_ling.stok_has_capacity(stok)):
            Cvr_Dict['cover_parse_cap'].append(1)
            Cvr_Dict['cover_tokens'].append(stok.text_with_ws)
            Cvr_Dict['cover_tokens_stok_subsets'].append([i_stok])
            Cvr_Dict['cover_tokens_cap'].append(1)
        else:
            Cvr_Dict['cover_parse_cap'].append(0)
            if(i_stok==0):
                # start_on_non_encoding_ele = True
                Cvr_Dict['cover_tokens'].append(stok.text_with_ws)
                Cvr_Dict['cover_tokens_stok_subsets'].append([i_stok])
                Cvr_Dict['cover_tokens_cap'].append(0)
            else:
                Cvr_Dict['cover_tokens'][-1] = Cvr_Dict['cover_tokens'][-1] + stok.text_with_ws
                Cvr_Dict['cover_tokens_stok_subsets'][-1].append(i_stok)

    # for now cover_list duplicatates cover_tokens 
    for tok in Cvr_Dict['cover_tokens']: Cvr_Dict['cover_list'].append(tok)

    return Cvr_Dict


def build_cover(
        cover_input: cover_input_type_alias,
        variant_cfg_input: dict | None=None,
    ) -> PM_Cover:

    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    if(pm_helpers.is_string(cover_input)):
        cover_text = cover_input
    elif(pm_helpers.is_list_of_strings(cover_input)):
        cover_text = " ".join(cover_input)
    elif(pm_helpers.is_list_of_spacyTokens(cover_input)):
        cover_text = str(cover_input[0].doc)
    elif(pm_helpers.is_spacyDoc(cover_input)):
        cover_text = str(cover_input)
    else:
        cover_text = None
        
    Cvr_Dict: PM_Cover = tokenize_cover(
        cover_text=cover_text,
        variant_cfg_input=variant_cfg
    )
    
    # logger.debug(f'Cvr_Dict:\n{pm_pformat(Cvr_Dict)}')

    return Cvr_Dict



# ---------------------------------------------------------------------------- #
# paramorse.__init__
# ---------------------------------------------------------------------------- #

"""
Paramorse: ParaLinguistic Morse code, a human super language.

WIP public API:
- encode(...) -> str
- decode(...) -> str
- configure_logs(...) -> None

"""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _pkg_version

# RE-EXPORT PUBLIC API

from paramorse.core.transform import encode, decode
from paramorse.utils.logging import configure_logs

try:
    __version__ = _pkg_version("paramorse")
except PackageNotFoundError:
    __version__ = "0.unknown"

__all__ = [
    "encode",
    "decode",
    "configure_logs",
    "__version__",
]



# ---------------------------------------------------------------------------- #
# paramorse.core.package_list
# ---------------------------------------------------------------------------- #

from __future__ import annotations

import random
import copy

from typing import Any 
from typing import TypedDict

from typing import Optional

import logging

from paramorse.utils.render import pm_pformat # noqa: F401 

import paramorse.core.linguistics as pm_ling
from paramorse.core.config import ParaMorseConfig

logger = logging.getLogger(__name__)


PAD_LEFT_PARA_SPACING_ALL_TOKS = True
PUNCT_AWARE_PAD_LEFT_PARA_SPACING_ALL_TOKS = True
TRIM_RIGHT_PARA_SPACING_LAST_TOK = False

# ---------------------------------------------------------------------------- #

class PM_Package(TypedDict):
    package_list: list[str]
    package_str: str


_PM_PACKAGES: dict[str, PM_Package] = {
    'empty': PM_Package(
        package_list = [],
        package_str = "",
    )
}


# ---------------------------------------------------------------------------- #

def parse_package_input(
        package_input: str,
        variant_cfg_input: dict[str, Any] | None=None,
    ) -> dict:
    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    dd_para_map: dict = variant_cfg['dd_para_map']

    snlp = pm_ling.load_spacy_model(variant_cfg["spacy_model"])
    package_parse_sdoc = snlp(package_input)
    package_parse_stoks = list(package_parse_sdoc)

    package_parse_list_lower_strip = [
        stok.text.lower().strip() for stok in package_parse_stoks
        ]

    cover_list = []
    para_flat = []

    for pkg_str in package_parse_list_lower_strip:
        if(pkg_str in dd_para_map.values()):
            para_flat.append(pkg_str)
        else:
            cover_list.append(pkg_str)

    pay_config = {
        'pay_format':'string_list',
        'pay_kind':'para'
    }
    
    parse_pkg_dict = {
        'package_input': package_input,
        'pay_config': pay_config,
        'para_flat': para_flat,
        'cover_list': cover_list,
    }

    return parse_pkg_dict


# ---------------------------------------------------------------------------- #


def build_package(
        cover_tokens: list[str],
        cover_tokens_cap: list,
        payload_para_flat: list[str],
        preamble_len: int = 0,
        variant_cfg_input: dict | None=None,
    ) -> PM_Package:

    # variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    Pkg_Dict: PM_Package = copy.deepcopy(_PM_PACKAGES['empty'])

    # allow_adjacent = variant_cfg['allow_adjacent'] # for now ignore

    cover_len = len(cover_tokens)
    if(preamble_len > cover_len):
        logger.warning("truncating preamble_len")
        preamble_len = cover_len
    if(preamble_len > 0):
        for i in range(preamble_len):
            cover_tokens_cap[i]=0

    if(len(payload_para_flat) > 0):
        package_list = splice_payload_into_cover(
            cover_tokens=cover_tokens,
            cover_tokens_cap=cover_tokens_cap,
            payload_para_flat=payload_para_flat
        )
    else:
        package_list = cover_tokens

    package_str = "".join(package_list)

    Pkg_Dict = {
        'package_list': package_list,
        'package_str': package_str
    }

    # logger.debug(f'Pkg_Dict:\n{pm_pformat(Pkg_Dict)}')

    return Pkg_Dict


def splice_payload_into_cover(
        cover_tokens: list[str],
        cover_tokens_cap: list[int],
        payload_para_flat: list[str],
        seed: Optional[int] = None,
    ) -> list[str]:

    if len(cover_tokens) != len(cover_tokens_cap):
        raise ValueError("cover_tokens and cover_tokens_cap must be the same length.")
    if any(c not in (0, 1) for c in cover_tokens_cap):
        raise ValueError("cover_tokens_cap must contain only 0 or 1.")

    cap1_indices = [i for i, c in enumerate(cover_tokens_cap) if c == 1]
    cap1_tokens = [cover_tokens[i] for i in cap1_indices] 

    cover_len = len(cover_tokens)
    payload_len = len(payload_para_flat)

    # number of slots available around the tokens with cap==1
    k = len(cap1_tokens) + 1 
    m = payload_len
    # playload paras can be 1 greater than the number of tokens with capacity

    # enforce non-adjacency capacity
    if m > k:
        logger.warning("truncating payload")
        payload_para_flat = payload_para_flat[:k]
        m = k

    rng = random.Random(seed) if seed is not None else random

    # randomly choose slots
    k_slots_range = range(k)
    k_slots = sorted(rng.sample(k_slots_range, m)) if m > 0 else []
    # k_slots = sorted(rng.choices(k_slots_range, k=m)) if m > 0 else [] # for allow adj

    k_slots_para = {}
    for k_slot, para in zip(k_slots, payload_para_flat):
        k_slots_para[k_slot] = para

    package_list: list[str] = []

    if not cap1_indices:
        # if no cap==1 tokens, then only one k_slot (k == 1).
        # place para (if any) for k_slot 0 at beginning
        # rest is the entirety of the cover
        para_at_pos0 = k_slots_para.get(0)
        if para_at_pos0 is not None:
            package_list.append(para_at_pos0)
        package_list.extend(cover_tokens)
        return package_list

    # therefore at least one cap==1 token
    first_cap1_idx = cap1_indices[0] 
    # walk the original cover tokens in order
    # keeping track of the cap==1 tokens passed 
    # at each cover token, look for the "after" k_slot index 
    # see if it exists and insert para token
    curr_cap1_pos = 0 # position within cap1_tokens (0..len-1)
    head = None
    head_is_para = False
    for i in range(cover_len):

        # ZERO-TH PARA TOKEN
        # if/when at the first cap==1,  check if there is k_slot for pos0
        # insert the "before first" k_slot
        if i == first_cap1_idx:
            if(i>0): first_token_is_cap0 = True
            else: first_token_is_cap0 = False
            para_at_pos0 = k_slots_para.get(0)
            if para_at_pos0 is not None:
                if(first_token_is_cap0): # add any post processing
                    para_at_pos0 = edit_spacing_at_para_tok_start(
                        preceding_token=cover_tokens[i-1],
                        para_tok=para_at_pos0
                        )
                head = para_at_pos0
                head_is_para = True
                package_list.append(head)

        # COVER TOKENS
        # add cover token itself
        head = cover_tokens[i]
        head_is_para = False
        package_list.append(head)

        # ALL OTHER PARA TOKENS
        # if cover token has capacity and next k_slot follows immediately
        # insert the "after" k_slot para token 
        if cover_tokens_cap[i] == 1:
            # k_slot indices: after 0th cap1 is k_slot 1, etc.
            slot_after = curr_cap1_pos + 1
            para_after = k_slots_para.get(slot_after)
            if para_after is not None:
                para_after = edit_spacing_at_para_tok_start(
                    preceding_token=cover_tokens[i],
                    para_tok=para_after
                    )
                head = para_after
                head_is_para = True
                package_list.append(head)
            curr_cap1_pos += 1

    if(head_is_para and TRIM_RIGHT_PARA_SPACING_LAST_TOK):
        package_list[-1] = edit_spacing_at_para_tok_end(para_tok=package_list[-1])

    # logger.debug(f'package_list: {pm_pformat(package_list)}')

    return package_list


def edit_spacing_at_para_tok_start(preceding_token: str, para_tok: str):

    # assumes para is of default form, eg. "uh " or "um "
    proc_para = para_tok
    if(PAD_LEFT_PARA_SPACING_ALL_TOKS):
        if(PUNCT_AWARE_PAD_LEFT_PARA_SPACING_ALL_TOKS):
            if(not pm_ling.last_char_is_space(pkg_substring=preceding_token)
               and not pm_ling.last_char_is_no_space_punct(pkg_substring=preceding_token)):
                proc_para = " " + para_tok
        else:
            if(not pm_ling.last_char_is_space(pkg_substring=preceding_token)):
                proc_para = " " + para_tok
                
    return proc_para


def edit_spacing_at_para_tok_end(para_tok: str):

    proc_para = para_tok
    if(TRIM_RIGHT_PARA_SPACING_LAST_TOK):
        proc_para = proc_para[:-1]

    return proc_para


# ---------------------------------------------------------------------------- #
# paramorse.utils.logging
# ---------------------------------------------------------------------------- #

from __future__ import annotations

import logging
import os
from typing import Optional


_DEFAULT_LOG_LEVEL = logging.DEBUG # 10
# _DEFAULT_LOG_LEVEL = logging.INFO # 20
# _DEFAULT_LOG_LEVEL = logging.WARNING # 30
# _DEFAULT_LOG_LEVEL = logging.ERROR # 40
# _DEFAULT_LOG_LEVEL = logging.CRITICAL # 50

_DEFAULT_LOG_ASCTIME = "%H:%M:%S"

class ParaMorseLogFormatter(logging.Formatter):

    # ANSI color codes
    GREY = "\x1b[38;20m" # "\x1b[38;21m"
    YELLOW = "\x1b[33;20m" # \x1b[33;21m"
    RED = "\x1b[31;20m" # "\x1b[31;21m"
    BOLD_RED = "\x1b[31;1m"
    CYAN = "\x1b[0;36m"
    PURPLE = "\x1b[38;2;182;90;190m"

    RESET = "\x1b[0m" 
    
    FORMAT_START = "| "
    FORMAT_MID = "%(levelname)s"

    # FORMAT_END = " |\n| %(pathname)s\n| %(module)s:%(lineno)d -- %(funcName)s --| %(message)s"
    FORMAT_END = " -- %(filename)s:%(lineno)d -- %(funcName)s --| %(message)s"
    # FORMAT_END = " -- %(asctime)s | %(filename)s:%(lineno)d -- %(funcName)s --| %(message)s"

    _PLAIN = FORMAT_START + FORMAT_MID + FORMAT_END

    FORMATS_COLORED = {
        logging.DEBUG: FORMAT_START + PURPLE + FORMAT_MID + RESET + FORMAT_END,
        logging.INFO: FORMAT_START + CYAN + FORMAT_MID + RESET + FORMAT_END,
        logging.WARNING: FORMAT_START + YELLOW + FORMAT_MID + RESET + FORMAT_END,
        logging.ERROR: FORMAT_START + RED + FORMAT_MID + RESET + FORMAT_END,
        logging.CRITICAL: FORMAT_START + BOLD_RED + FORMAT_MID + RESET + FORMAT_END
    }
    
    def __init__(self, *, datefmt: Optional[str] = _DEFAULT_LOG_ASCTIME, use_color: bool = True) -> None:
        super().__init__(datefmt=datefmt)
        self.datefmt = datefmt
        self.use_color = use_color
    
    def format(self, record: logging.LogRecord) -> str:
        if self.use_color:
            log_fmt = self.FORMATS_COLORED.get(record.levelno, self._PLAIN)
        else:
            log_fmt = self._PLAIN
        return logging.Formatter(log_fmt, datefmt=self.datefmt).format(record)



def configure_logs(
        level: int | None = _DEFAULT_LOG_LEVEL,
        *,
        datefmt: str = "%H:%M:%S",
        use_color: bool | None = None,
        force: bool = False,
        silence_numba: bool = True,
        numba_level: int = logging.WARNING,
    ) -> None:
    """
    Configure root logging with a colored formatter.
    If handlers already exist and force=False, do not replace them.
    Args:
        level:
            root level passed to the formatter
        datefmt:
            time format passed to the formatter
        use_color:
            force color on/off
        force:
            if True, replace existing handlers
        silence_numba & numba_level:
            if True, set numba logger level to `numba_level`.
    """
    root = logging.getLogger()

    if silence_numba:
        logging.getLogger("numba").setLevel(numba_level)

    # decide whether to change handlers
    if root.handlers and not force:
        # adjust level even if we keep existing handlers
        if level is not None:
            root.setLevel(level)
        return
    
    effective_level = logging.INFO if level is None else level

    handler = logging.StreamHandler()
    if use_color is None:
        use_color = _auto_use_color(handler.stream)
    handler.setFormatter(ParaMorseLogFormatter(datefmt=datefmt, use_color=use_color))

    logging.basicConfig(level=effective_level, handlers=[handler], force=force)


def _auto_use_color(stream) -> bool:
    # respect NO_COLOR (https://no-color.org/) and only color TTYs
    if os.environ.get("NO_COLOR"):
        return False
    try:
        return hasattr(stream, "isatty") and stream.isatty()
    except Exception:
        return False



# ---------------------------------------------------------------------------- #
# paramorse.core.config
# ---------------------------------------------------------------------------- #

from __future__ import annotations

import copy

from typing import Any, Union
from collections.abc import Mapping, MutableMapping

import logging

logger = logging.getLogger(__name__)


# ---------------------------------------------------------------------------- #


_SEPAR_DOT_TOKEN = "um"
_SEPAR_DASH_TOKEN = "uh"
_SEPAR_PARA_LETSEP = "okay"
# SEPAR_PARA_WORDSEP = ""
_SEPAR_PARA_WORDSEP = "alright"
# SEPAR_WORDSEP_OVERWRITE = False
_SEPAR_WORDSEP_OVERWRITE = True

_SEPAR_DD_PARA_MAP: dict[str, str] = {
    ".": _SEPAR_DOT_TOKEN,
    "-": _SEPAR_DASH_TOKEN,
    " ": _SEPAR_PARA_LETSEP,
    "/": _SEPAR_PARA_WORDSEP,
}

_MORSE_ALPHABET_SYMBOL_DD: dict[str, str] = {
    "A": ".-",
    "B": "-...",
    "C": "-.-.",
    "D": "-..",
    "E": ".",
    "F": "..-.",
    "G": "--.",
    "H": "....",
    "I": "..",
    "J": ".---",
    "K": "-.-",
    "L": ".-..",
    "M": "--",
    "N": "-.",
    "O": "---",
    "P": ".--.",
    "Q": "--.-",
    "R": ".-.",
    "S": "...",
    "T": "-",
    "U": "..-",
    "V": "...-",
    "W": ".--",
    "X": "-..-",
    "Y": "-.--",
    "Z": "--..",
    "0": "-----",
    "1": ".----",
    "2": "..---",
    "3": "...--",
    "4": "....-",
    "5": ".....",
    "6": "-....",
    "7": "--...",
    "8": "---..",
    "9": "----.",
    ".": ".-.-.-",
    ",": "--..--",
    "?": "..--..",
    "'": ".----.",
    "!": "-.-.--",
    "/": "-..-.",
    "(": "-.--.",
    ")": "-.--.-",
    "&": ".-...",
    ":": "---...",
    ";": "-.-.-.",
    "=": "-...-",
    "+": ".-.-.",
    "-": "-....-",
    "_": "..--.-",
    '"': ".-..-.",
    "$": "...-..-",
    "@": ".--.-.",
}

# build morse alphabet maps
_MORSE_ALPHABET_SYMBOL_PARA: dict[str, list[str]] = {}
for sym, dd in _MORSE_ALPHABET_SYMBOL_DD.items():
    para = []
    for ele in list(dd):
        para.append(_SEPAR_DD_PARA_MAP[ele])
    _MORSE_ALPHABET_SYMBOL_PARA[sym] = para


_MORSE_ALPHABET_DOTDASH_SYM: dict[str, str] = {
    dd: sym for sym, dd in _MORSE_ALPHABET_SYMBOL_DD.items()
}


SEPAR_CONFIG: dict[str, Union[str, bool, int, list, dict]] = {

    # USERy
    # paralanguage settings
    # para token values
    "dot_token": _SEPAR_DOT_TOKEN,
    "dash_token": _SEPAR_DASH_TOKEN,
    "letter_separator": _SEPAR_PARA_LETSEP,  # paralanguage marking end of payload letter (empty = disabled)
    "word_separator": _SEPAR_PARA_WORDSEP,  # paralanguage marking end of payload word (empty = disabled)
    # para token operations
    "word_sep_overwrite": _SEPAR_WORDSEP_OVERWRITE,  # overwrite letter separator with word separator (combines spacing, if applicable)
    "allow_adjacent": False,

    # SYSTEMy
    # para token map
    "dd_para_map": _SEPAR_DD_PARA_MAP,
    # morse payload
    # morse payload alphabet
    "symbol_to_dd": _MORSE_ALPHABET_SYMBOL_DD,
    # morse payload alphabet maps
    "dd_to_symbol": _MORSE_ALPHABET_DOTDASH_SYM,
    "symbol_to_para": _MORSE_ALPHABET_SYMBOL_PARA,
    # spaCy settings
    "spacy_model": "en_core_web_sm",
    "language": "en-US",

}


# ---------------------------------------------------------------------------- #


class ParaMorseConfig(dict):
    """
    A dictionary-like configuration whose `.update()` keeps:
      `symbol_to_para`
      `dd_para_map`
    in sync.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(copy.deepcopy(SEPAR_CONFIG))
        if args or kwargs:
            # force a rebuild so the initial overrides are coherent
            self.update(*args, _force_rebuild=True, **kwargs)

    def update(self, *args: Any, _force_rebuild: bool = False, **kwargs: Any) -> None:
        """
        Works like `dict.update`, but refreshes the derived maps
        when any of the dependent keys change.
        """

        # normalize positional & keyword inputs into a single mapping
        update_params_dict: dict[str, Any] = {}

        if args:
            if len(args) > 1:
                raise TypeError("update expected at most 1 positional argument")

            if len(args) == 1:
                arg = args[0]
                if isinstance(arg, Mapping):
                    update_params_dict.update(arg)
                else:
                    # assume iterable of (k, v) pairs
                    for k, v in arg:
                        update_params_dict[k] = v

        update_params_dict.update(kwargs)
        needs_rebuild = _force_rebuild or any(
            k in _FORCE_RBLD_KEYS and self.get(k) != v
            for k, v in update_params_dict.items()
        )
        super().update(update_params_dict)

        if needs_rebuild:
            _rebuild_para_maps(self)
            # logger.debug("rebuilt symbol_to_para & dd_para_map due to config change.")

    # convenience: clear() + replace everything in one call
    def replace(self, **new_cfg: Any) -> None:
        """
        Convenience clear() with replace all settings in one call
        while preserving dict identity.
        """
        self.clear()
        super().update(copy.deepcopy(SEPAR_CONFIG))
        self.update(new_cfg, _force_rebuild=True)


# ---------------------------------------------------------------------------- #

_FORCE_RBLD_KEYS: set = {
    "dot_token",
    "dash_token",
    "letter_separator",
    "word_separator",
}

_DEPENDANT_KEYS: set = {
    "dd_para_map",
    "symbol_to_para"
}


def _rebuild_para_maps(variant_cfg: MutableMapping[str, Any]) -> None:
    """
    Mutates `variant_cfg` in-place, refreshing:
      `symbol_to_para`
      `dd_para_map`
    """

    dot_para = variant_cfg["dot_token"]
    dash_para = variant_cfg["dash_token"]
    letter_sep_para = variant_cfg["letter_separator"]
    word_sep_para = variant_cfg["word_separator"]
    symbol_to_dd: dict[str, str] = variant_cfg["symbol_to_dd"]

    # _DEPENDANT_KEYS updates

    # dd_para_map
    variant_cfg["dd_para_map"] = {
        ".": dot_para,
        "-": dash_para,
        " ": letter_sep_para,
        "/": word_sep_para,
    }

    # symbol_to_para
    new_symbol_para = {}
    # for sym, dd in variant_cfg['symbol_to_dd'].items():
    for sym, dd in symbol_to_dd.items():
        para = dd
        for key, val in variant_cfg["dd_para_map"].items():
            para = [c.replace(key, val) for c in para]
        new_symbol_para[sym] = para
    variant_cfg["symbol_to_para"] = new_symbol_para



# ---------------------------------------------------------------------------- #
# paramorse.utils.render
# ---------------------------------------------------------------------------- #

import os
import logging
import textwrap
import difflib

from typing import Any 
from pprint import PrettyPrinter

from paramorse.utils.paths import get_data_dirs

# IPython is optional
# try to import and if unavailable, fall back at call sites

try:
    from IPython.display import Markdown as _IPy_Markdown
    from IPython.display import display as _IPy_display
    _HAS_IPYTHON = True
except Exception:
    _IPy_Markdown = None
    _IPy_display = None
    _HAS_IPYTHON = False


logger = logging.getLogger(__name__)

_DATADIRS: dict[str, dict]=get_data_dirs()

_pp = PrettyPrinter(
    indent=2,
    width=80,
    compact=True,
    # compact=False,
    # sort_dicts=True,
    sort_dicts=False,
)

_MAX_TABLE_WIDTH = 140

_MAX_PRINT_LEN = 650




# PRINT & FORMAT

def trunc_str(text: str, max_char: None | int=_MAX_PRINT_LEN):

    truncated = False
    
    if max_char is not None and len(text) > max_char:
        text = text[:max_char]
        truncated = True

    if truncated and text:
        text = text + " ..."

    return text


def pm_pformat(input: Any):
    return _pp.pformat(input)


def pm_pprint(input: Any):
    return _pp.pprint(input)




# TEXT COMPARE

def text_compare_html_diff(
        expected_string: str,
        actual_string: str,
        compare_filepath: str | None=None,
        test_data_subdir: str = "",
        test_compare_filename: str = 'text_compare.html',
    ) -> bool:

    # make into lists
    expected_lines = expected_string.splitlines()
    actual_lines = actual_string.splitlines()

    # compare
    # text_differ = difflib.Differ()
    # diff = text_differ.compare(expected_string, actual_string)
   
    html_diff = difflib.HtmlDiff(wrapcolumn=60).make_file(
        expected_lines,
        actual_lines
        )
    
    if(compare_filepath is not None):
        resolve_compare_path = os.path.abspath(compare_filepath)
    else:
        resolve_compare_path = f"{_DATADIRS['paths_string']['test']}/{test_data_subdir}/{test_compare_filename}"

    with open(f'{resolve_compare_path}', 'w') as f:        
        f.write(html_diff)

    if os.path.isfile(resolve_compare_path):
        return True
    else:
        return False




# MARKDOWN TABLES
# TODO types

def md_table_build(
    rows,
    columns,
    justify=None,
    # *,
    col_min_widths=None,
    table_max_width : None | int=_MAX_TABLE_WIDTH,
    target="terminal",  # "terminal" or "notebook"
):
    """
    Build a Markdown table string (pipes).

    Parameters
    ----------
    rows : list[dict|sequence]
        - Preferred (concise): list of sequences, each aligned by position to `columns`.
          Example: rows = [['config', config], ['cover', cover]]
        - Back-compat: list of dicts mapping column -> value.
    columns : list[str]
        Column headers (order defines column order).
    justify : list[str] | None
        One per column in {'left','center','right'} (or 'l','c','r'). Defaults to all 'left'.
    col_min_widths : None | int | list[int] | dict[str,int]
        Minimum content width per column (hard lower bound). If dict, keys are headers.
    table_max_width : int | None
        If int, maximum overall row width (including pipes/spaces). Cells are **wrapped** to fit.
        If None, no cap is enforced.
    target : "notebook" | "terminal"
        - notebook: HTML-escaped, line breaks become <br>
        - terminal: raw Markdown, wrapped cells emit **additional table rows**, padded to width
    """

    # ---------------- helpers (target-agnostic) ----------------
    def _norm_justify(just):
        if just is None:
            return ["unspec"] * len(columns)
        if len(just) != len(columns):
            raise ValueError("justify must have the same length as columns")
        out = []
        for j in just:
            jj = str(j).lower()
            if jj in ("u","unspec","unspecified"): out.append("unspec")
            elif jj in ("l","left"): out.append("left")
            elif jj in ("c","center","centre","middle"): out.append("center")
            elif jj in ("r","right"): out.append("right")
            else:
                raise ValueError(f"invalid justify value: {j!r}; use unspec/left/center/right")
        return out

    def _norm_min_widths(mcw):
        if mcw is None:
            return [0] * len(columns)
        if isinstance(mcw, int):
            return [max(0, mcw)] * len(columns)
        if isinstance(mcw, list):
            if len(mcw) != len(columns):
                raise ValueError("col_min_widths list must match len(columns)")
            return [max(0, int(v)) for v in mcw]
        if isinstance(mcw, dict):
            return [max(0, int(mcw.get(c, 0))) for c in columns]
        raise ValueError("col_min_widths must be None, int, list[int], or dict[str,int]")

    def _wrap_text(raw, width):
        """Wrap raw text to width; returns '\\n'-joined lines (no escaping)."""
        if width <= 0:
            return ""
        s = "" if raw is None else str(raw)
        s = s.replace("\r\n", "\n").replace("\r", "\n")
        out = []
        for part in s.split("\n"):
            wrapped = textwrap.wrap(
                part,
                width=width,
                break_long_words=True,
                break_on_hyphens=True,
                replace_whitespace=False,
                drop_whitespace=False,
            )
            if not wrapped:
                out.append("")
            else:
                out.extend(wrapped)
        return "\n".join(out)

    # Normalize rows to a list-of-lists aligned with columns (backward compatible)
    def _normalize_rows(rows, columns):
        norm = []
        for r_idx, row in enumerate(rows):
            if isinstance(row, dict):
                norm.append([row.get(c, "") for c in columns])
            else:
                try:
                    seq = list(row)
                except TypeError:
                    raise ValueError(f"Row {r_idx} is neither dict nor sequence")
                if len(seq) < len(columns):
                    seq = seq + [""] * (len(columns) - len(seq))
                elif len(seq) > len(columns):
                    # Stricter here to avoid silent misalignment
                    raise ValueError(
                        f"Row {r_idx} has {len(seq)} cells; expected {len(columns)}"
                    )
                norm.append(seq)
        return norm

    # ---------------- target-specific helpers ----------------
    tgt = str(target).lower()
    if tgt not in ("notebook", "terminal"):
        raise ValueError("target must be 'notebook' or 'terminal'")

    def _visible_len_notebook(x):
        s = "" if x is None else str(x)
        return len(s.replace("\n", " "))

    def _visible_len_terminal(x):
        s = "" if x is None else str(x)
        s = s.replace("|", "\\|")
        return len(s.replace("\n", " "))

    def _escape_for_notebook(s):
        # Escape &, <, >; escape '|'; newline -> <br>
        s = s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
        s = s.replace("|", "\\|")
        return s.replace("\n", "<br>")

    def _escape_for_terminal(s):
        # Only escape pipes; keep real newlines
        return s.replace("|", "\\|")

    def _align_token(j):
        return {"unspec": "---", "left": ":---", "center": ":---:", "right": "---:"}[j]

    def _sep_token_terminal(width, j):
        w = max(3, int(width))
        if j == "unspec":
            return "-" + "-"*(w-1)
        if j == "left":
            return ":" + "-"*(w-1)
        elif j == "right":
            return "-"*(w-1) + ":"
        else:
            return ":" + "-"*max(1, w-2) + ":"

    def _pad_align(s, width, j):
        n = len(s)
        if n >= width:
            return s
        if j == "right":
            return " "*(width - n) + s
        if j == "center":
            left = (width - n) // 2
            right = width - n - left
            return " "*left + s + " "*right
        return s + " "*(width - n)

    # ---------------- normalize inputs ----------------
    N = len(columns)
    justify = _norm_justify(justify)
    min_widths = [max(1, w) for w in _norm_min_widths(col_min_widths)]  # >=1

    # Normalize rows to matrix (list of lists)
    rows_mat = _normalize_rows(rows, columns)

    # choose vlen based on target
    vlen = _visible_len_notebook if tgt == "notebook" else _visible_len_terminal

    # measure content
    header_vis = [vlen(h) for h in columns]
    body_max = [0] * N
    for row in rows_mat:
        for i in range(N):
            body_max[i] = max(body_max[i], vlen(row[i]))

    # initial widths
    widths = [max(header_vis[i], body_max[i], min_widths[i]) for i in range(N)]

    # enforce overall cap if set
    if table_max_width is not None:
        non_content_extra = 3 * N + 1  # "| " + " | ".join(cells) + " |"
        content_budget = max(0, table_max_width - non_content_extra)
        # terminal: don't shrink below header_vis (single-line header)
        low_bounds = (
            min_widths[:] if tgt == "notebook"
            else [max(min_widths[i], header_vis[i]) for i in range(N)]
        )
        over = sum(widths) - content_budget
        if over > 0 and N > 0:
            while over > 0:
                reducible = [i for i in range(N) if widths[i] > low_bounds[i]]
                if not reducible:
                    break
                share = max(1, over // len(reducible))
                changed = False
                for i in reducible:
                    dec = min(share, widths[i] - low_bounds[i])
                    if dec > 0:
                        widths[i] -= dec
                        over -= dec
                        changed = True
                    if over <= 0:
                        break
                if not changed:
                    for i in reducible:
                        if widths[i] > low_bounds[i] and over > 0:
                            widths[i] -= 1
                            over -= 1
                    if over > 0 and all(widths[i] == low_bounds[i] for i in range(N)):
                        break

    # ---------------- build ----------------
    if tgt == "notebook":
        # Header (wrapped; HTML-escaped; <br> for newlines)
        hdr_cells = []
        for i, c in enumerate(columns):
            wrapped = _wrap_text(c, widths[i])
            hdr_cells.append(_escape_for_notebook(wrapped))
        header = "| " + " | ".join(hdr_cells) + " |"

        # Alignment row
        sep = "| " + " | ".join(_align_token(j) for j in justify) + " |"

        # Body
        body_lines = []
        for r in rows_mat:
            cells = []
            for i in range(N):
                wrapped = _wrap_text(r[i], widths[i])
                cells.append(_escape_for_notebook(wrapped))
            body_lines.append("| " + " | ".join(cells) + " |")
            # body_lines.append("| ```" + "``` | ```".join(cells) + "``` |")

        return "\n".join([header, sep] + body_lines)

    # ----- terminal path -----
    # Single-line header (replace header newlines with spaces)
    term_hdr_cells = []
    for i, c in enumerate(columns):
        raw = "" if c is None else str(c)
        raw = raw.replace("\r\n", "\n").replace("\r", "\n").replace("\n", " ")
        cell = _escape_for_terminal(raw)
        term_hdr_cells.append(_pad_align(cell, widths[i], justify[i]))
    header = "| " + " | ".join(term_hdr_cells) + " |"

    # Alignment row sized to final widths
    sep = "| " + " | ".join(_sep_token_terminal(widths[i], justify[i]) for i in range(N)) + " |"

    # Body rows: wrap each cell to width; emit as multiple physical rows; pad each segment
    body_lines = []
    for r in rows_mat:
        col_lines = []
        max_h = 0
        for i in range(N):
            esc = _escape_for_terminal("" if r[i] is None else str(r[i]))
            wrapped = _wrap_text(esc, widths[i])
            lines = wrapped.split("\n") if wrapped else [""]
            max_h = max(max_h, len(lines))
            col_lines.append(lines)

        for k in range(max_h):
            row_cells = []
            for i in range(N):
                seg = col_lines[i][k] if k < len(col_lines[i]) else ""
                row_cells.append(_pad_align(seg, widths[i], justify[i]))
            body_lines.append("| " + " | ".join(row_cells) + " |")

    return "\n".join([header, sep] + body_lines)


def md_table(
    rows,
    columns = ['Item', 'Value'],
    justify = ['left', 'left'],
    # *,
    col_min_widths = [13, 13],
    table_max_width : None | int=_MAX_TABLE_WIDTH,
):
    
    if not _HAS_IPYTHON:
        # Fallback: produce terminal/raw Markdown string
        return md_table_str(
            rows=rows, columns=columns, justify=justify,
            col_min_widths=col_min_widths,
            table_max_width=table_max_width,
        )
    
    # IPython present: build notebook-flavored table and wrap in Markdown
    if _IPy_Markdown is not None:
        """Notebook-friendly: returns IPython.display.Markdown"""
        return _IPy_Markdown(md_table_build(
            rows=rows, columns=columns, justify=justify,
            col_min_widths=col_min_widths,
            table_max_width=table_max_width,
            target="notebook",
        ))


def md_table_str(
    rows,
    columns = ['Item', 'Value'],
    justify = ['unspec', 'unspec'],
    # *,
    col_min_widths = [13, 13],
    table_max_width : None | int=_MAX_TABLE_WIDTH,
):
    """Terminal/raw: returns a Markdown string suitable for printing"""
    return md_table_build(
        rows, columns, justify,
        col_min_widths=col_min_widths,
        table_max_width=table_max_width,
        target="terminal",
    )


def display_md_table(
    rows,
    columns = ['Item', 'Value'],
    justify = ['left', 'left'],
    # *,
    col_min_widths = [13, 13],
    table_max_width : None | int=_MAX_TABLE_WIDTH,
):
    if not _HAS_IPYTHON:
        # Fallback: print the terminal-friendly Markdown string
        return print_md_table_str(
            rows=rows, columns=columns, justify=justify,
            col_min_widths=col_min_widths,
            table_max_width=table_max_width,
        )
    # IPython present
    if _IPy_display is not None:
        return _IPy_display(md_table(
            rows=rows, columns=columns, justify=justify,
            col_min_widths=col_min_widths,
            table_max_width=table_max_width,
        ))


def print_md_table_str(
        rows,
        columns = ['Item', 'Value'],
        justify = ['unspec', 'unspec'],
        # *,
        col_min_widths = [13, 13],
        table_max_width : None | int=_MAX_TABLE_WIDTH,
    ):

    return print(md_table_str(
        rows=rows, columns=columns, justify=justify,
        col_min_widths=col_min_widths,
        table_max_width=table_max_width,
    ))


# ---------------------------------------------------------------------------- #
# paramorse.core.transform
# ---------------------------------------------------------------------------- #

from __future__ import annotations

from typing import Any
from typing import TypedDict

import json
import logging 
import copy
from paramorse.utils.render import pm_pformat # noqa: F401 

from paramorse.utils.paths import data_path_resolver
from paramorse.utils.paths import get_data_dirs

import paramorse.utils.helpers as pm_helpers

from paramorse.core.config import ParaMorseConfig

# import paramorse.core as pm_core
import paramorse.core.linguistics as pm_ling
import paramorse.core.payload as pm_payload
import paramorse.core.cover as pm_cover
import paramorse.core.package as pm_package

logger = logging.getLogger(__name__)

_DATADIRS: dict[str, dict] = get_data_dirs()


# ---------------------------------------------------------------------------- #

class PM_SeparEnc(TypedDict):
    package_list: list[str]
    package_str: str
    package_dict: pm_package.PM_Package
    cover_dict: pm_cover.PM_Cover
    payload_dict: pm_payload.PM_Payload
    variant_cfg: ParaMorseConfig


class PM_SeparDec(TypedDict):
    pay_para: list # para_flat
    pay_morse: str
    pay_text: str
    variant_cfg: ParaMorseConfig


_PM_SEPAR_ENCS: dict[str, PM_SeparEnc] = {
    'empty': PM_SeparEnc(
        package_list = [],
        package_str = "",
        package_dict = copy.deepcopy(pm_package._PM_PACKAGES['empty']),
        cover_dict = copy.deepcopy(pm_cover._PM_COVERS['empty']),
        payload_dict = copy.deepcopy(pm_payload._PM_PAYLOADS['empty']),
        variant_cfg = ParaMorseConfig(),
    )
}


_PM_SEPAR_DECS: dict[str, PM_SeparDec] = {
    'empty': PM_SeparDec(
        pay_para = [],  # para_flat list
        pay_morse = "", # dd string
        pay_text = "", # sym string
        variant_cfg = ParaMorseConfig(),
    )
}



# ---------------------------------------------------------------------------- #

def encode(
        cover: str,
        payload: str,
        config: dict | None=None,
        output_path: str | None=None,
    ) -> str:

    enc_out_dict = None
    if(output_path is not None):
        if(isinstance(output_path, str)):
            enc_out_dict = {'data_file_path': output_path}
        else:
            logger.error('encode output_path must be a string type (str)')

    SeparEnc_Dict = encode_separ(
        cover_input = cover,
        payload_input = payload,
        variant_cfg_input = config,
        enc_out_dict = enc_out_dict,
    )
    package = SeparEnc_Dict['package_str']

    return package


def encode_separ(
        cover_input: str | None = None,
        payload_input: pm_payload.payload_input_type_alias | None = None, # detects payload type and format
        variant_cfg_input: dict[str, Any] | None = None,
        enc_out_dict: dict[str,Any] | None = None,
    ) -> PM_SeparEnc:

    SeparEnc_Dict: PM_SeparEnc = copy.deepcopy(_PM_SEPAR_ENCS['empty'])
    
    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()
    SeparEnc_Dict['variant_cfg'] = variant_cfg


    # COVER

    if( (cover_input is None or payload_input is None)
       or (cover_input == "") ):
        return SeparEnc_Dict

    cover_dict = pm_cover.build_cover(
        cover_input=cover_input,
        variant_cfg_input=variant_cfg
        )
    
    cover_tokens = cover_dict['cover_tokens']
    cover_parse_list = cover_dict['cover_parse_list']


    # PAYLOAD

    payload_dict = pm_payload.build_payload(
        payload_input=payload_input,
        variant_cfg_input=variant_cfg
        )


    # PACKAGE

    # para_toks_in_cover_set = pm_core.linguistics.para_toks_in_cover(
    para_toks_in_cover_set = pm_ling.para_toks_in_cover(
        cover_parse_list=cover_parse_list,
        payload_para_flat=payload_dict['para_flat']
        )
    if (len(para_toks_in_cover_set)>0):
        raise ValueError(f"cover_parse_list contains para tokens: {para_toks_in_cover_set}")

    package_dict = pm_package.build_package(
        cover_tokens=cover_tokens
        ,cover_tokens_cap=cover_dict['cover_tokens_cap']
        ,payload_para_flat=payload_dict['para_flat_rs']
        ,variant_cfg_input=variant_cfg
    )

    package_list = package_dict['package_list']
    package_str = package_dict['package_str']

    SeparEnc_Dict = {
        'package_list': package_list, # TODO duplicated data
        'package_str': package_str,
        'package_dict': package_dict, 
        'cover_dict': cover_dict,
        'payload_dict': payload_dict,
        'variant_cfg': variant_cfg
    }

    if(enc_out_dict is not None):
        encode_save_to_file(
            enc_dict=SeparEnc_Dict,
            enc_out_dict=enc_out_dict,
            variant_cfg_input=variant_cfg_input,
            variant_cfg=variant_cfg,
            )
        
    return SeparEnc_Dict


def encode_save_to_file(
        enc_dict: PM_SeparEnc,
        enc_out_dict: dict = {},
        variant_cfg_input: dict | None=None,
        variant_cfg: dict | None=None,
    ) -> bool:

    encode_saved = False

    save_dict: dict[str, Any]
    if pm_helpers.is_nonempty_dict(enc_out_dict):
        save_dict = enc_out_dict
    else:
        save_dict = {

            # "data_file_path": None,

            # "data_dir_name": None,
            "data_dir_name": _DATADIRS['paths_string']['encoded_pkg'],

            # "data_item_name": "example_item",
            # "data_item_name": "decoded_package",

            # "data_item_subname": None,
            # "data_item_subname": "example_subitem",

            # 'data_item_id': "example_id",

            # "data_file_suffix": ".txt"
        }
    path_dict = data_path_resolver(
        **save_dict
        )
    # logger.debug(f'path_dict:\n{pm_pformat(path_dict)}')

    if (pm_helpers.is_nonempty_dict(enc_dict)
        and pm_helpers.is_nonempty_dict(path_dict)):

        enc_out_filepath = path_dict['file_path_obj']
        enc_out_dir = path_dict['dir_path_obj']

        if(path_dict['resolve_source'] == 'data_file_path'):
            save_filepath = enc_out_filepath
        else:
            save_filename = "enc_pkg_text.txt"
            save_filepath = enc_out_dir / save_filename

        logger.debug(f"saving output to:\n{save_filepath}")
        enc_data_str = enc_dict['package_str']
        with open(save_filepath, 'w', encoding='utf-8') as f:
            f.write(enc_data_str)

        if(path_dict['resolve_source'] == 'data_file_path'):
            enc_out_dir_stem = path_dict['file_path_obj_stem']
            save_filepath = enc_out_dir / (enc_out_dir_stem + "_meta.json")
        else:
            save_filename = "enc_pkg_meta.json"
            save_filepath = enc_out_dir / save_filename

        enc_meta_dict = {
            'variant_cfg_input' : variant_cfg_input,
            'variant_cfg' : variant_cfg
        }
        logger.debug(f"saving output to:\n{save_filepath}")
        with open(save_filepath, "w") as f:
            json.dump(enc_meta_dict, f, indent=4)

        encode_saved = True    

    return encode_saved


# ---------------------------------------------------------------------------- #

def decode(
        package: str,
        config: dict | None=None,
        output_path: str | None=None,
    ) -> str:

    dec_out_dict = None
    if(output_path is not None):
        if(isinstance(output_path, str)):
            dec_out_dict = {'data_file_path': output_path}
        else:
            logger.error('decode output_path must be a string type (str)')

    SeparDec_Dict = decode_separ(
        package = package,
        variant_cfg_input = config,
        dec_out_dict = dec_out_dict,
    )
    payload = SeparDec_Dict['pay_text']

    return payload


def decode_separ(
        package: str,
        variant_cfg_input: dict[str, Any] | None = None,
        dec_out_dict: dict | None = None,
    ) -> PM_SeparDec:

    SeparDec_Dict: PM_SeparDec = copy.deepcopy(_PM_SEPAR_DECS['empty'])

    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()
        
    SeparDec_Dict['variant_cfg'] = variant_cfg

    # right now spacy tokens "stoken" are approximation of variant tokens
    parsed_dict = pm_package.parse_package_input(
        package_input=package,
        variant_cfg_input=variant_cfg
    )
    # logger.debug(f'parsed_dict:\n{parsed_dict}')

    para_flat = parsed_dict['para_flat']
    pay_config = parsed_dict['pay_config']

    payload_dict = pm_payload.build_payload(
        para_flat
        ,pay_config
        ,variant_cfg_input=variant_cfg
        )

    decoded_morse = payload_dict['dd']
    decoded_text = payload_dict['sym']

    SeparDec_Dict = {
        "pay_para": para_flat,
        "pay_morse": decoded_morse,
        "pay_text": decoded_text,
        'variant_cfg': variant_cfg,
        }
       
    if(dec_out_dict is not None):
        decode_save_to_file(
            dec_dict=SeparDec_Dict, 
            dec_out_dict=dec_out_dict, 
            variant_cfg_input=variant_cfg_input,
            variant_cfg=variant_cfg,
            )

    return SeparDec_Dict


def decode_save_to_file(
        dec_dict: PM_SeparDec,
        dec_out_dict: dict = {},
        variant_cfg_input: dict | None=None,
        variant_cfg: dict | None=None,
    ) -> bool:

    decode_saved = False

    save_dict: dict[str, Any]
    if pm_helpers.is_nonempty_dict(dec_out_dict):
        save_dict = dec_out_dict
    else:
        save_dict = {
            # "data_file_path": None,

            # "data_dir_name": None,
            "data_dir_name": _DATADIRS['paths_string']['decoded_pkg'],

            # "data_item_name": "example_item",
            # "data_item_name": "decoded_package",

            # "data_item_subname": None,
            # "data_item_subname": "example_subitem",

            # 'data_item_id': "example_id",

            # "data_file_suffix": ".txt"

        }
    path_dict = data_path_resolver(
        **save_dict
        )
    # logger.debug(f'path_dict:\n{pm_pformat(path_dict)}')

    if (pm_helpers.is_nonempty_dict(dec_dict)
        and pm_helpers.is_nonempty_dict(path_dict)):

        dec_out_filepath = path_dict['file_path_obj']
        dec_out_dir = path_dict['dir_path_obj'] 

        if(path_dict['resolve_source'] == 'data_file_path'):
            save_filepath = dec_out_filepath
        else:
            save_filename = "dec_pkg.json"
            save_filepath = dec_out_dir / save_filename
        
        dec_data_dict = {
            'pay_para' : dec_dict['pay_para'],
            'pay_text' : dec_dict['pay_text'],
            'pay_morse' : dec_dict['pay_morse'],
        }
        logger.debug(f"saving output to:\n{save_filepath}")
        with open(save_filepath, "w") as f:
            json.dump(dec_data_dict, f, indent=4)

        if(path_dict['resolve_source'] == 'data_file_path'):
            enc_out_dir_stem = path_dict['file_path_obj_stem']
            save_filepath = dec_out_dir / (enc_out_dir_stem + "_meta.json")
        else:
            save_filename = "dec_pkg_meta.json"
            save_filepath = dec_out_dir / save_filename
            
        dec_meta_dict = {
            'variant_cfg_input' : variant_cfg_input,
            'variant_cfg' : variant_cfg,
        }
        logger.debug(f"saving output to:\n{save_filepath}")
        with open(save_filepath, "w") as f:
            json.dump(dec_meta_dict, f, indent=4)

        decode_saved = True

    return decode_saved





# ---------------------------------------------------------------------------- #
# paramorse.core.linguistics
# ---------------------------------------------------------------------------- #

from __future__ import annotations

import functools

import logging
import spacy

from spacy.tokens.token import Token as spacyToken
from spacy.language import Language as spacyLanguage


logger = logging.getLogger(__name__)

# ---------------------------------------------------------------------------- #
# sets of linguistic objects 

_CONTRACTIONS = {
    "n't",
    "'s",
    "'d",
    "'ll",
    "'re",
    "'ve",
    "'m",
    }


# _PUNCTUATION = {

#     ".",
#     "?",
#     "!",

#     ",",
#     ":",
#     ";",

#     "}",
#     "{",
#     "]",
#     "[",
#     ")",
#     "(",
#     ">",
#     "<",
    
# }

_PUNCTS_NO_PAD_LEFT_SPACE = {

    # ".",
    # "?",
    # "!",

    # ",",
    # ":",
    # ";",

    # "}",
    "{",
    # "]",
    "[",
    # ")",
    "(",
    # ">",
    "<",

}

# _PUNCTS_NEXT_TO_SPACE = {
#     ws
#     for p in _PUNCTUATION
#     # for ws in (f"{p} ")
#     for ws in (f" {p}", f"{p} ", f" {p} ")
# }

# _SPACINGS_NON_STANDARD = {
#     "\t",
#     "\n",
#     "\r",
#     # " ",
# }

# _SPACINGS = _SPACINGS_NON_STANDARD.union({" "})


# ---------------------------------------------------------------------------- #
# spaCy

@functools.lru_cache(maxsize=1)
def load_spacy_model(model_name: str) -> spacyLanguage:
    try:
        return spacy.load(model_name)
    except OSError:
        spacy.cli.download(model_name) # type: ignore
        return spacy.load(model_name)


# ---------------------------------------------------------------------------- #
# paramorse package creation


# cover-playoad token overlap 
def para_toks_in_cover(
        cover_parse_list: list,
        payload_para_flat: list
    ) -> set:
    cover_toks_set = set(cover_parse_list)
    para_toks_set = set(payload_para_flat)
    toks_intersection = para_toks_set.intersection(cover_toks_set)
    return toks_intersection


# cover token has capacity to carry morse para
def stok_has_capacity(stok: spacyToken) -> bool:

    stok_text = stok.text

    stok_has_capacity = not (

        False

        or stok.is_punct
        or stok.is_bracket
        # or stok_text in _PUNCTUATION
        # or stok_text in _PUNCTS_NEXT_TO_SPACE

        or stok.is_space
        # or stok_text in _SPACINGS_NON_STANDARD
        # or stok_text in _SPACINGS

        or stok_text in _CONTRACTIONS
        or stok_text.startswith("'")
    
    )

    return stok_has_capacity


# ---------------------------------------------------------------------------- #


def last_char_is_space(pkg_substring: str) -> bool:
    last_char = pkg_substring[-1]
    if(last_char.isspace()): is_space = True
    else: is_space = False
    return is_space


def last_char_is_no_space_punct(pkg_substring: str) -> bool:
    last_char = pkg_substring[-1]
    if(last_char in _PUNCTS_NO_PAD_LEFT_SPACE): is_no_space_punct = True
    else: is_no_space_punct = False
    return is_no_space_punct



# ---------------------------------------------------------------------------- #
# paramorse.core.generator
# ---------------------------------------------------------------------------- #

from __future__ import annotations

import functools
import random
import textwrap
import logging
from typing import Any

import functools
from importlib.resources import files
from pathlib import Path


from paramorse.utils.helpers import read_text_file

import paramorse.core.transform as pm_transform
import paramorse.core.payload as pm_payload
import paramorse.core.cover as pm_cover
from paramorse.core.config import ParaMorseConfig

logger = logging.getLogger(__name__)


_DEFAULT_WORDS_FILE = "words_1k.txt"

""" GENERATE COVERS """


def _read_packaged_text(package: str, relative_path: str) -> str:
    return (files(package) / relative_path).read_text(encoding="utf-8")

@functools.lru_cache(maxsize=5)
def load_wordlist(words_file: str = _DEFAULT_WORDS_FILE) -> list[str]:
   
    p = Path(words_file)
    if p.is_file():
        # user-supplied override by path
        words_input = p.read_text(encoding="utf-8")
    else:
        # packaged fallback
        words_input = _read_packaged_text("paramorse", f"data/{words_file}")

    return words_input.splitlines()


def rand_list_words(
        n_words: int = 10,
        trailing_space: bool = False
    ) -> list:

    wordlist = load_wordlist()
    list_words = random.sample(wordlist, n_words)

    if(trailing_space):
        for i in range(len(list_words)):
            list_words[i] = list_words[i] + " "

    return list_words


def rand_two_word_list(
        n_word_pairs: int = 5,
        mode: str = "alternating",
        preamble: int = 0,
        trailing_space: bool = False
    ) -> list:
    
    if(n_word_pairs<0): n_word_pairs=0
    if(preamble<0): preamble=0

    word_1 = None
    word_2 = None
    max_word_len = 8

    wordlist = load_wordlist()

    while (word_1 is None
            or len(word_1) > max_word_len):
            word_1 = random.sample(wordlist, 1)[0]
    while (word_2 is None
            or word_2 == word_1
            or len(word_2) > max_word_len):
            word_2 = random.sample(wordlist, 1)[0]

    word_bank = [word_1, word_2]

    list_words = []

    if(preamble>0):
        for i in range(0, preamble, 1):
            if(i%2==0):
                list_words.append(word_1)
            else:
                list_words.append(word_2)

    if(mode=="alternating"):    
        for _ in range(0, n_word_pairs, 1):
            list_words.append(word_1)
            list_words.append(word_2)
    elif(mode=="random"):
        for _ in range(0, n_word_pairs, 1):
            list_words.append(random.choice(word_bank))
            list_words.append(random.choice(word_bank))

    if(trailing_space):
         for i in range(len(list_words)):
              list_words[i] = list_words[i] + " "

    return list_words


def intro_exercise_cover() -> dict:
    
    text_string = '''
    This message is in ParaMorse, a super language variant of oral English.
    ParaMorse maps the dots and dashes of Morse code onto paralinguistic filler sounds.
    What symbol has been embedded in this message?
    '''

    text_string = textwrap.dedent(text_string)
    text_string = text_string[1:-1]

    cover_dict = pm_cover.build_cover(text_string)
    
    exercise_dict = {
        'text_string': text_string,
        'cover_dict': cover_dict,
    }

    return exercise_dict




""" GENERATE PAYLOADS """


def rand_payload_sym(
        n_symbols: int=2,
        spaced: bool=False,
        variant_cfg_input: dict[str, Any] | None = None,
    ) -> str:

    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()
    
    # default alphabet indices:
    # 0-26 letters
    # 26-37 numbers
    # 37-54 punctuation & special chars
    N_THRU_LETTERS = 26
    # N_THRU_NUMBERS = 36
    payload_alphabet_n = N_THRU_LETTERS

    rand_source = list(variant_cfg['symbol_to_dd'])[0:payload_alphabet_n]
    rand_symbols = []
    for _ in range(0, n_symbols):
        rand_char = random.choice(rand_source)
        rand_symbols.append(rand_char)
        if(spaced):
            rand_symbols.append(" ")

    rand_sym_string = "".join(rand_symbols)

    return rand_sym_string


def get_payload_list_dicts(
        idx_lower: int = 0,
        idx_upper: int = 26,
        payload_type: str = 'sym',
        variant_cfg_input: dict[str, Any] | None = None,
    ) -> dict:
    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    # default alphabet indices:
    # 0-26 letters
    # 26-37 numbers
    # 37-54 punctuation & special chars
    # N_THRU_LETTERS = 26
    # N_THRU_NUMBERS = 36

    if(payload_type == 'sym'):
        list_source = list(variant_cfg['symbol_to_dd'])
        symbol_list = list_source[idx_lower:idx_upper]
        payload_list = symbol_list
    else:
        # payload_list = ['hello', 'world']
        payload_list = ['hello world']

    payload_dict_list = {}
    for pay in payload_list:
        payload_dict_list[f'pay_{pay}'] = pm_payload.build_payload(payload_input=pay)
    
    return payload_dict_list




""" GENERATE PACKAGES """


def rand_encode_separ(
        cover_input: str | None = None,
        payload_input: pm_payload.payload_input_type_alias | None = None, # detects payload type and format
        variant_cfg_input: dict[str, Any] | None = None,
        enc_out_dict: dict[str,Any] | None = None,
    ) -> pm_transform.PM_SeparEnc:

    variant_cfg = ParaMorseConfig(**variant_cfg_input) if variant_cfg_input else ParaMorseConfig()

    if(cover_input is None):
        word_len = 30
        list_words = rand_two_word_list(n_word_pairs=word_len)
        cover_input = " ".join(list_words)

    if(payload_input is None):
        letter_len = 2
        payload_input = rand_payload_sym(
            n_symbols=letter_len,
            spaced=True,
            variant_cfg_input=variant_cfg,
            )

    encode_dict = pm_transform.encode_separ(
        cover_input=cover_input,
        payload_input=payload_input,
        variant_cfg_input=variant_cfg,
        enc_out_dict=enc_out_dict,
    )

    return encode_dict
